var __create = Object.create;
var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __getProtoOf = Object.getPrototypeOf;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toESM = (mod, isNodeMode, target) => (target = mod != null ? __create(__getProtoOf(mod)) : {}, __copyProps(
  // If the importer is in node compatibility mode or this is not an ESM
  // file that has been converted to a CommonJS file using a Babel-
  // compatible transform (i.e. "__esModule" has not been set), then set
  // "default" to the CommonJS "module.exports" for node compatibility.
  isNodeMode || !mod || !mod.__esModule ? __defProp(target, "default", { value: mod, enumerable: true }) : target,
  mod
));
var __toCommonJS = (mod) => __copyProps(__defProp({}, "__esModule", { value: true }), mod);
var query_exports = {};
__export(query_exports, {
  cacheMiddleware: () => cacheMiddleware,
  parseBuilder: () => parseBuilder,
  parseFieldAndAssociations: () => parseFieldAndAssociations,
  parseVariables: () => parseVariables,
  postProcess: () => postProcess,
  query: () => query,
  queryData: () => queryData
});
module.exports = __toCommonJS(query_exports);
var import_database = require("@nocobase/database");
var import_formatter = require("./formatter");
var import_koa_compose = __toESM(require("koa-compose"));
var import_utils = require("@nocobase/utils");
const getDB = (ctx, dataSource) => {
  const ds = ctx.app.dataSourceManager.dataSources.get(dataSource);
  return ds == null ? void 0 : ds.collectionManager.db;
};
const postProcess = async (ctx, next) => {
  const { data, fieldMap } = ctx.action.params.values;
  ctx.body = data.map((record) => {
    Object.entries(record).forEach(([key, value]) => {
      if (!value) {
        return;
      }
      const { type } = fieldMap[key] || {};
      switch (type) {
        case "bigInt":
        case "integer":
        case "float":
        case "double":
          record[key] = Number(value);
          break;
      }
    });
    return record;
  });
  await next();
};
const queryData = async (ctx, next) => {
  const { dataSource, collection, queryParams, fieldMap } = ctx.action.params.values;
  const db = getDB(ctx, dataSource) || ctx.db;
  const model = db.getModel(collection);
  const data = await model.findAll(queryParams);
  ctx.action.params.values = {
    data,
    fieldMap
  };
  await next();
};
const parseBuilder = async (ctx, next) => {
  const { dataSource, measures, dimensions, orders, include, where, limit } = ctx.action.params.values;
  const db = getDB(ctx, dataSource) || ctx.db;
  const { sequelize } = db;
  const attributes = [];
  const group = [];
  const order = [];
  const fieldMap = {};
  let hasAgg = false;
  measures.forEach((measure) => {
    const { field, aggregation, alias } = measure;
    const attribute = [];
    const col = sequelize.col(field);
    if (aggregation) {
      hasAgg = true;
      attribute.push(sequelize.fn(aggregation, col));
    } else {
      attribute.push(col);
    }
    if (alias) {
      attribute.push(alias);
    }
    attributes.push(attribute.length > 1 ? attribute : attribute[0]);
    fieldMap[alias || field] = measure;
  });
  dimensions.forEach((dimension) => {
    const { field, format, alias, type } = dimension;
    const attribute = [];
    const col = sequelize.col(field);
    if (format) {
      attribute.push((0, import_formatter.formatter)(sequelize, type, field, format));
    } else {
      attribute.push(col);
    }
    if (alias) {
      attribute.push(alias);
    }
    attributes.push(attribute.length > 1 ? attribute : attribute[0]);
    if (hasAgg) {
      group.push(attribute[0]);
    }
    fieldMap[alias || field] = dimension;
  });
  orders.forEach((item) => {
    const alias = sequelize.getQueryInterface().quoteIdentifier(item.alias);
    const name = hasAgg ? sequelize.literal(alias) : sequelize.col(item.field);
    order.push([name, item.order || "ASC"]);
  });
  ctx.action.params.values = {
    ...ctx.action.params.values,
    queryParams: {
      where,
      attributes,
      include,
      group,
      order,
      limit: limit || 2e3,
      subQuery: false,
      raw: true
    },
    fieldMap
  };
  await next();
};
const parseFieldAndAssociations = async (ctx, next) => {
  const {
    dataSource,
    collection: collectionName,
    measures,
    dimensions,
    orders,
    filter
  } = ctx.action.params.values;
  const db = getDB(ctx, dataSource) || ctx.db;
  const collection = db.getCollection(collectionName);
  const fields = collection.fields;
  const models = {};
  const parseField = (selected) => {
    var _a, _b, _c, _d;
    let target;
    let name;
    if (!Array.isArray(selected.field)) {
      name = selected.field;
    } else if (selected.field.length === 1) {
      name = selected.field[0];
    } else if (selected.field.length > 1) {
      [target, name] = selected.field;
    }
    const rawAttributes = collection.model.getAttributes();
    let field = ((_a = rawAttributes[name]) == null ? void 0 : _a.field) || name;
    let fieldType = (_b = fields.get(name)) == null ? void 0 : _b.type;
    if (target) {
      const targetField = fields.get(target);
      const targetCollection = db.getCollection(targetField.target);
      const targetFields = targetCollection.fields;
      fieldType = (_c = targetFields.get(name)) == null ? void 0 : _c.type;
      field = `${target}.${field}`;
      name = `${target}.${name}`;
      const targetType = (_d = fields.get(target)) == null ? void 0 : _d.type;
      if (!models[target]) {
        models[target] = { type: targetType };
      }
    } else {
      field = `${collectionName}.${field}`;
    }
    return {
      ...selected,
      field,
      name,
      type: fieldType,
      alias: selected.alias || name
    };
  };
  const parsedMeasures = (measures == null ? void 0 : measures.map(parseField)) || [];
  const parsedDimensions = (dimensions == null ? void 0 : dimensions.map(parseField)) || [];
  const parsedOrders = (orders == null ? void 0 : orders.map(parseField)) || [];
  const include = Object.entries(models).map(([target, { type }]) => ({
    association: target,
    attributes: [],
    ...type === "belongsToMany" ? { through: { attributes: [] } } : {}
  }));
  const filterParser = new import_database.FilterParser(filter, {
    collection
  });
  const { where, include: filterInclude } = filterParser.toSequelizeParams();
  const parsedFilterInclude = filterInclude == null ? void 0 : filterInclude.map((item) => {
    var _a;
    if (((_a = fields.get(item.association)) == null ? void 0 : _a.type) === "belongsToMany") {
      item.through = { attributes: [] };
    }
    return item;
  });
  ctx.action.params.values = {
    ...ctx.action.params.values,
    where,
    measures: parsedMeasures,
    dimensions: parsedDimensions,
    orders: parsedOrders,
    include: [...include, ...parsedFilterInclude || []]
  };
  await next();
};
const parseVariables = async (ctx, next) => {
  const { filter } = ctx.action.params.values;
  if (!filter) {
    return next();
  }
  const isNumeric = (str) => {
    if (typeof str === "number")
      return true;
    if (typeof str != "string")
      return false;
    return !isNaN(str) && !isNaN(parseFloat(str));
  };
  const getUser = () => {
    return async ({ fields }) => {
      var _a, _b;
      const userFields = fields.filter((f) => f && ctx.db.getFieldByPath("users." + f));
      (_a = ctx.logger) == null ? void 0 : _a.info("parse filter variables", { userFields, method: "parseVariables" });
      if (!ctx.state.currentUser) {
        return;
      }
      if (!userFields.length) {
        return;
      }
      const user = await ctx.db.getRepository("users").findOne({
        filterByTk: ctx.state.currentUser.id,
        fields: userFields
      });
      (_b = ctx.logger) == null ? void 0 : _b.info("parse filter variables", {
        $user: user == null ? void 0 : user.toJSON(),
        method: "parseVariables"
      });
      return user;
    };
  };
  ctx.action.params.values.filter = await (0, import_utils.parseFilter)(filter, {
    timezone: ctx.get("x-timezone"),
    now: (/* @__PURE__ */ new Date()).toISOString(),
    getField: (path) => {
      const fieldPath = path.split(".").filter((p) => !p.startsWith("$") && !isNumeric(p)).join(".");
      const { resourceName } = ctx.action;
      return ctx.db.getFieldByPath(`${resourceName}.${fieldPath}`);
    },
    vars: {
      $nDate: (0, import_utils.getDateVars)(),
      $user: getUser()
    }
  });
  await next();
};
const cacheMiddleware = async (ctx, next) => {
  const { uid, cache: cacheConfig, refresh } = ctx.action.params.values;
  const cache = ctx.app.cacheManager.getCache("data-visualization");
  const useCache = (cacheConfig == null ? void 0 : cacheConfig.enabled) && uid;
  if (useCache && !refresh) {
    const data = await cache.get(uid);
    if (data) {
      ctx.body = data;
      return;
    }
  }
  await next();
  if (useCache) {
    await cache.set(uid, ctx.body, (cacheConfig == null ? void 0 : cacheConfig.ttl) * 1e3);
  }
};
const checkPermission = (ctx, next) => {
  if (ctx.action.params.values === void 0) {
    ctx.action.params.values = ctx.action.params;
  }
  const { collection } = ctx.action.params.values;
  const roleName = ctx.state.currentRole || "anonymous";
  const can = ctx.app.acl.can({ role: roleName, resource: collection, action: "list" });
  if (!can && roleName !== "root") {
    ctx.throw(403, "No permissions");
  }
  return next();
};
const query = async (ctx, next) => {
  try {
    await (0, import_koa_compose.default)([
      checkPermission,
      cacheMiddleware,
      parseVariables,
      parseFieldAndAssociations,
      parseBuilder,
      queryData,
      postProcess
    ])(ctx, next);
  } catch (err) {
    ctx.throw(500, err);
  }
};
// Annotate the CommonJS export names for ESM import in node:
0 && (module.exports = {
  cacheMiddleware,
  parseBuilder,
  parseFieldAndAssociations,
  parseVariables,
  postProcess,
  query,
  queryData
});
